/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 708 $ $LastChangedDate: 2002-08-30 10:56:28 +0200 (Fri, 30 Aug 2002) $
 *
 * Ksi - Interface to the Ksi intermediate language.
 */

type id;

/* Ksi type system. */
union ty {
	void Void;
	void Bool;
	int Int;		/* int is size */
	int UInt;
	int Float;
	ty Ptr;			/* ptr to ... */
	ty Reference;		/* reference to ... */
	*(ty, <ty>list) FuncPtr;
				/* (return_type, parameters) */
	void Ellipsis;		/* `...' to be used only in function parms */
	id Struct;
	id Union;
	id MacroType;		/* MacroType("c_int") -> .c_int */
	id TyRef;		/* (tyref foobar) */
	void Poly;
}

/* Given a string, return Ksi identifier.  */
id make_id(string);

/* Return type of signed integer of the same size as native pointer.  */
ty ptroff();

/* Ditto, but unsigned.  */
ty uptroff();

/* Attributes that associated with functions, variables and types.
   At the very moment only Extern, Local and Public are used.  However I
   hope to use NoReturn and MallocLike some day...  */
union attr {
	int Aligned;
	void Packed;
	void Local		/* this is the default anyway... */;
	void Public;
	void Extern;
	void SuppressDebug;
	void ConstFunc;
	void PureFunc;
	void MallocLike;
	void NoReturn;
	void Artificial;
	void Weak;
	void OneOnly;
	void Comdat;
	string Section;
	string ExportName;
	id AliasSet;
}

union binop {
	void Eq;		/* == */
	void Ne;		/* != */
	void Ge;		/* >= */
	void Gt;		/* > */
	void Le;		/* <= */
	void Lt;		/* < */
	void Plus;		/* + */
	void Minus;		/* - */
	void Mult;		/* * */
	void RDiv;		/* float / float */
	void ExactDiv;	/* division to use when no modulo is expected, like
			 * division of pointer difference by object size
			 * in C. */
	void RoundDiv;	/* round result */
	void FloorDiv;	/* round result down */
	void CeilDiv;	/* round result up */
	void TruncDiv;	/* truncate result */

	/* corresponding modulus operations */
	void RoundMod;
	void FloorMod;
	void CeilMod;
	void TruncMod;

	void BitAnd;	/* & */
	void BitIOr;	/* | */
	void BitAndTc;	/* sth strange */
	void BitXor;	/* ^ */
	void Max;
	void Min;
	void LRotate;
	void LShift;	/* << */
	void RRotate;
	void RShift;	/* >> */

	/* ++/--, first operand is l-value to modify,
	 * second operand is how much to inc/decrement by. */
	void PostInc;
	void PreInc;
	void PostDec;
	void PreDec;

	void TruthAndIf;	/* && */
	void TruthOrIf;		/* || */
	void TruthAnd;		/* similar to && but not lazy */
	void TruthOr;		/* similar to || but not lazy */
	void TruthXor;		/* same, but ^ is used instead of | or & */
}

union unop {
	void Negate;		/* -X */
	void BitNot;		/* ~X */
	void TruthNot;		/* !X */
	void Abs;
	void Addr;		/* &X */
	void RAddr;		/* As Addr, but return reference. */
	void Ffs;		/* position of first bit set */
	void NonLvalue;		/* Hm... this is strange... */
}

union expr {
	/* (const ...) term */
	*(ty, Int64::t) IntConst;
	*(ty, float) FloatConst;
	bool BoolConst;
	string StringConst;

	id EnumRef;

	/* null constant of type ty */
	ty Null;

	/* Single term of similar name: */
	*(ty, id) Var;
	*(ty, id) StaticVar;
	expr Bind;		/* create scope */
	id Ref;			/* reference variable by name */
	id FuncRef;		/* get address of named function */
	*(expr, expr) Modify;	/* expr = expr */
	*(expr, id) Component;	/* expr.id */
	*(expr, id) Arrow;	/* expr->string, shortcut */

	/* this is shortcut to (return (modify (result) expr))
	 * expr can be Nop, (return (nil)) is generated then */
	expr Return;

	ty SizeOf;			/* Returns (const .sizetype ...). */
	<expr>list Compound;		/* expr1, expr2, ..., exprn */
	expr Loop;			/* while (1) expr; */
	void Nop;
	void Result;
	expr Indirect;			/* *expr */
	*(expr, expr, expr) Cond;	/* expr ? expr : expr */
	/* call named function, passing parameters */
	*(id, <expr>list) Call;
	/* call function that is guaranteed not to have any side-effects */
	*(id, <expr>list) FCall;
	*(expr, <expr>list) PtrCall;
	*(expr, <expr>list) PtrFCall;
	*(id, <expr>list) BuiltIn;
	*(id, <expr>list) FBuiltIn;
	*(ty, expr) Convert;
	*(id, expr) Block;		/* create named block */
	id ExitBlock;			/* jump to end of named block */
	*(binop, expr, expr) Binary;
	*(unop, expr) Unary;
	*(expr, expr) TryFinally;		/* try e1 finally e2 */
	/* annotate expression with line number information. */
	*(int, expr) Line;
	/* annotate output Ksi code with comment */
	*(string, expr) Comment;
}

union decl {
	*(<attr>list, ty, id, <ty>list) FuncProto;
	*(<attr>list, ty, id, <*(ty, id)>list, expr) Func;
	string TopComment;		/* output comment */
	id StructDefProto;
	*(<attr>list, id, <*(ty, id)>list) StructDef;
	id UnionDefProto;
	*(<attr>list, id, <*(ty, id)>list) UnionDef;
	*(<attr>list, id, ty) TyAlias;
	*(<attr>list, id) AbstractType;
	<*(id, int)>list EnumDef;
	*(<attr>list, ty, id) Global;
	string InputFile;
}

/* Print header of Ksi file to given channel.  */
void output_header(Stdio::FILE out);

/* Print one declaration to output channel.  */
void output_decl(Stdio::FILE out, decl d);

/* Finish writing Ksi file.  */
void output_trailer(Stdio::FILE out);

/* Set to true to get ugly-looking Ksi code (generation of it is however
   much faster).  Ergh... it doesn't work yet.  */
bool fast_writer;
