/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 703 $ $LastChangedDate: 2002-08-30 01:19:46 +0200 (Fri, 30 Aug 2002) $
 *
 * Linker - Link checks, gont-link-aux.ksi generation, linking programs and
 * making libraries.
 */
 
open Ast;
open Util;
open Fileutil;
open Xobjects;
open Linker;

'a load_xml(cmdline c, string filename, *('a (Xml::node)) f)
{
    try {
	filename = find_in_path(c.search_path, filename);
    } with {
    case Not_found:
	report_error("cannot find file: `" + filename + "'");
    }
    
    try {
        return f(Xml::root_node(Xml::parse_file(filename, true)));
    } with {
    case Invalid_argument(_):
	report_error("XML object in file `" + filename + "' is malformed");
	raise Choke;
    }
}

xgo load_xgo(cmdline c, string filename)
{
    return load_xml(c, filename, unserialize_xgo);
}

xga load_xga(cmdline c, string filename)
{
    return load_xml(c, filename, unserialize_xga);
}

bool ends_with(string what, string suf)
{
    let wl = String::length(what) in
    let sl = String::length(suf) in
    return wl >= sl && String::sub(what, wl - sl, sl) == suf;
}

void linkcheck(<xgo>list objs, bool relaxed)
{
    int err_cnt = 0;
    
    xgo find_module(<xgo>list lst, id m)
    {
	return List::find(fun bool (xgo x) (x.xgo_name == m), lst);
    }

    bool needs_ini(xgo x)
    {
	return x.xgo_init_routine != "" || x.xgo_fini_routine != "";
    }

    void unweak(xgo x)
    {
	if (x.xgo_weak) {
	    x.xgo_weak = false;
	    void unweak_deps(id n)
	    {
		try {
		    unweak(find_module(objs, n));
		} with {
		case Not_found: skip;
		}
	    }

	    List::iter(unweak_deps, x.xgo_references);
	    List::iter(unweak_deps, x.xgo_uses);
	}
    }

    void maybe_unweak(xgo x)
    {
	if (!x.xgo_weak) {
	    x.xgo_weak = true;
	    unweak(x);
	}
    }

    void check_lst(<xgo>list lst)
    {
	switch lst {
	case Cons(x, rest):
	    <xgo>list before = rest;
	    xgo current = x;
	    bool check_ini = true;
	
	    void check_dependent(id name)
	    {
		try {
		    find_module(before, name);
		    // if found, then it's ok
		} with {
		case Not_found:
		    try {
			// if not found, try in all modules
			xgo dep = find_module(objs, name);
			if (check_ini && 
			    (needs_ini(dep) || needs_ini(current))) {
			    report_error(current.xgo_name + " needs proper " +
					 "initialization/finalization of " +
					 name + " before it is used");
			    err_cnt++;
			}
		    } with {
		    case Not_found:
			if (!relaxed) {
			    err_cnt++;
			    report_error(current.xgo_name + " needs " +
					 name + " which is not to be found");
			}
		    }
		}
	    }

	    if (!current.xgo_weak) {
		List::iter(check_dependent, current.xgo_uses);
		check_ini = false;
		List::iter(check_dependent, current.xgo_references);
	    }

	    check_lst(rest);
	
	case Nil:
	    skip;
	}
    }

    if (!relaxed)
	List::iter(maybe_unweak, objs);

    check_lst(objs);

    if (err_cnt > 0)
	raise Choke;
}

void mklib(cmdline c)
{
    <xgo>list objs = Nil;
    int err_cnt = 0;
    
    void load(string name)
    {
	if (!ends_with(name, ".xgo")) {
	    report_error("pathname `" + name + "' does not end with .xgo");
	    err_cnt++;
	} else {
	    try {
		objs = Cons(load_xgo(c, name), objs);
	    } with {
	    case Choke:
		err_cnt++;
	    }
	}
    }

    List::iter(load, c.input_files);

    if (List::length(objs) == 0) {
	report_error("cowardly refusing to create empty archive");
	err_cnt++;
    }

    string libname = "";
    
    switch split_filename(basename(c.output_file)) {
    case (name, "xga"):
	libname = name;
	
    case (_, _):
	report_error("pathname `" + c.output_file + "' does not end with .xga");
	err_cnt++;
    }
    
    if (err_cnt > 0)
	raise Choke;
  
    linkcheck(objs, true);
    
    xga x = ({
	xga_name = libname,
	xga_modules = List::rev(objs),
	xga_link_flags = c.link_flags
    });

    save_xga(x, c.output_file);

    // run ar?
}

void make_link_aux(cmdline c, <xgo>list objs)
{
    <int>Hashtbl::t exns = Hashtbl::create(123);
    int exn_no = 1;
    Stdio::FILE f = Stdio::stdout;
    <Ksi::expr>list exns_inits = Nil;
    <Ksi::expr>list ctor_fncs = Nil;
    <Ksi::expr>list init_fncs = Nil;
    <Ksi::expr>list fini_fncs = Nil;

    Ksi::expr make_call(string name, bool push_proto)
    {
	let n = Ksi::make_id(name) in
	let p = Ksi::FuncProto(List::cons1(Ksi::Extern), Ksi::Void, n, 
			       Nil) in 
	{
	    if (push_proto)
		Ksi::output_decl(f, p);
	    return Ksi::Call(n, Nil);
	}
	
    }
    
    void process_xgo(xgo x)
    {
	void add_one_exn(id i)
	{
	    if (!Hashtbl::mem(exns, i))
		Hashtbl::add(exns, i, exn_no++);
	}

	List::iter(add_one_exn, x.xgo_used_exns);

	if (x.xgo_ctor_routine != "")
	    ctor_fncs = Cons(make_call(x.xgo_ctor_routine, true),
				   ctor_fncs);

	if (x.xgo_init_routine != "")
	    init_fncs = Cons(make_call(x.xgo_init_routine, true),
				   init_fncs);

	if (x.xgo_fini_routine != "")
	    fini_fncs = Cons(make_call(x.xgo_fini_routine, true), 
				   fini_fncs);
	
	let xgo_linked = 
	         Ksi::Global(List::cons1(Ksi::Public), 
			     Ksi::ptroff(), 
			     Ksi::make_id(x.xgo_name + "$$xgo_linked$")) in
	Ksi::output_decl(f, xgo_linked);
    }

    void exn_init(string name, int no)
    {
	let n = Ksi::make_id(name) in
	let t = Ksi::ptroff() in
	let v = Int64::of_int(no) in
	let d = Ksi::Global(List::cons1(Ksi::Public), t, n) in
	let i = Ksi::Modify(Ksi::Ref(n), Ksi::IntConst(t, v)) in {
	    Ksi::output_decl(f, d);
	    exns_inits = Cons(i, exns_inits);
	}
    }

    void push_function(bool is_public, string name, <Ksi::expr>list body)
    {
	<Ksi::attr>list attrs;

	if (is_public)
	    attrs = List::cons1(Ksi::Public);
	else
	    attrs = List::cons1(Ksi::Local);

	let d = Ksi::Func(attrs, Ksi::Void, Ksi::make_id(name), Nil,
			  Ksi::Bind(Ksi::Compound(body))) in
	{
	    Ksi::output_decl(f, d);
	}
    }

    f = Stdio::fopen(c.gont_link_aux + ".ksi", "w");
    Ksi::output_header(f);

    List::iter(process_xgo, objs);
    Hashtbl::iter(exns, exn_init);

    push_function(true, "gont_init_modules", 
		  List::concat(List::concat(exns_inits, ctor_fncs), init_fncs));
    push_function(true, "gont_fini_modules", List::rev(fini_fncs));

    _ ptr = Ksi::MacroType(Ksi::make_id("ptr"));
    _ c_int = Ksi::MacroType(Ksi::make_id("c_int"));
    
    let reg_proto = Ksi::FuncProto(List::cons1(Ksi::Extern), 
				   Ksi::Void,
				   Ksi::make_id("gont_register_exit_routine"),
				   List::cons2(Ksi::FuncPtr(Ksi::Void, Nil),
					       ptr)) in
    let sav_proto = Ksi::FuncProto(List::cons1(Ksi::Extern), 
				   Ksi::Void,
				   Ksi::make_id("gont_set_argv"),
				   [c_int, ptr]) in
    {
	Ksi::output_decl(f, reg_proto);
	Ksi::output_decl(f, sav_proto);
    }

    let argc = Ksi::make_id("argc") in
    let argv = Ksi::make_id("argv") in
    let fini_modules = Ksi::FuncRef(Ksi::make_id("gont_fini_modules")) in
    let null_value = Ksi::Null(Ksi::Ptr(Ksi::Void)) in
    let set_argv = Ksi::Call(Ksi::make_id("gont_set_argv"), 
			     List::cons2(Ksi::Ref(argc), Ksi::Ref(argv))) in
    let atexit = Ksi::Call(Ksi::make_id("gont_register_exit_routine"), 
		           List::cons2(fini_modules, null_value)) in
    let main_hd = List::cons4(make_call("gont_init_runtime", true),
			      set_argv,
			      atexit, 
			      make_call("gont_init_modules", false)) in
    let main_tl = List::cons2(make_call("gont_run_exit_routines", true),
			      make_call("gont_exit", true)) in
    let main_body = Ksi::Bind(Ksi::Compound(List::concat(main_hd, main_tl))) in
    let main_decl = Ksi::Func([Ksi::Public], 
			      c_int,
			      Ksi::make_id("main"), 
			      [(c_int, argc), (ptr, argv)],
			      main_body) in
    {
	Ksi::output_decl(f, main_decl);
    }
    
    Ksi::output_trailer(f);
    Stdio::fclose(f);
}

*(string, <string>list) link(cmdline c)
{
    <xgo>list objs = Nil;
    <string>list cobjs = Nil;
    string link_flags = "";
    int err_cnt = 0;

    void add_library(string file, bool weak)
    {
	xga x = load_xga(c, file);
	void add_object(xgo o)
	{
	    o.xgo_weak = weak; 
	    objs = Cons(o, objs);
	}
	
	List::iter(add_object, x.xga_modules);

	if (x.xga_link_flags != "")
	    link_flags = link_flags + " " + x.xga_link_flags;
    }
    
    string find_file(string name)
    {
	try {
	    return find_in_path(c.search_path, name);
	} with {
	case Not_found:
	    report_error("could not find file `" + name + "'");
	    err_cnt++;
	    raise Choke;
	}
    }
    
    void handle_input_file(string file)
    {
	switch split_filename(file) {
	case (b, "xgo"):
	    objs = Cons(load_xgo(c, find_file(file)), objs);
	    cobjs = Cons(b + ".o", cobjs);
	    
	case (b, "xga"):
	    add_library(find_file(file), true);
	    cobjs = Cons("libgont" + b + ".a", cobjs);
	    
	case (b, "xga:force"):
	    add_library(find_file(b + ".xga"), false);
	    cobjs = Cons("libgont" + b + ".a", cobjs);
	
	case (_, "o"):
	    cobjs = Cons(file, cobjs);

	case (_, "a"):
	    cobjs = Cons(file, cobjs);

	case (_, "so"):
	    cobjs = Cons(file, cobjs);

	case (_, _):
	    report_error("don't know what to do with `" + file + "'");
	    err_cnt++;
	}
    }

    List::iter(handle_input_file, c.input_files);
    
    linkcheck(objs, false);
    
    cobjs = List::map(find_file, cobjs);
    
    if (err_cnt > 0)
	raise Choke;

    make_link_aux(c, objs);
	
    if (c.link_flags != "")
	link_flags = link_flags + " " + c.link_flags;

    cobjs = Cons(c.gont_link_aux + ".o", cobjs);
    
    return (link_flags, cobjs);
}

// vim:sts=4:sw=4
