/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 703 $ $LastChangedDate: 2002-08-30 01:19:46 +0200 (Fri, 30 Aug 2002) $
 *
 * Main - Driver program for Gont compiler.
 *
 * We are going to call linker and such from here in future. Now we only
 * call Gont::compile_interface or Gont::compile_implementation.
 */
 
open Gont;
open Fileutil;
open Util;

struct cmdline {
    <string>list input_files;
    string output_file;
    string ksi_compiler;    // -ksicc
    string ksi_flags;	    // -ksi
    string link_flags;	    // -ld
    string prefix;	    // -prefix

    <string>list lib_path;	    // -L
    <string>list module_path;    // -I
    
    bool just_make_ksi;	    // -S
    bool just_compile;	    // -c
    bool mkarch_mode;	    // -a
    int verbose;	    // -v
    bool ciface;	    // -ciface
}

bool did_some_compilation;

void bye()
{
    report_error("Use " + Array::get(Sys::argv, 0) + " -help to get more info");
    Sys::exit(1);
}

void die(string msg)
{
    report_error(msg);
    bye();
}

void just_one_compilation()
{
    if (did_some_compilation)
	die("sorry, can handle just one compilation at a time");
    did_some_compilation = true;
}

cmdline cli;

void verb(string msg)
{
    if (cli.verbose > 0)
	report_error(msg);
}

void sh(string cmd)
{
    verb("running: `" + cmd + "'");

    try {
	int ret = Sys::shell(cmd);
	if (ret != 0) {
	    report_error("command `" + cmd + 
			 "' finished with non-zero exit status");
	    Sys::exit(1);
	}
    } with {
    case Sys_error(m):
	report_error("failed to run `" + cmd + "': " + m);
	Sys::exit(1);
    }
}

void compile_ksi(string filename, string output)
{
    sh(cli.ksi_compiler + " " + cli.ksi_flags + 
	    " -c " + filename + 
	    " -o " + output);
}

string compile_file(string filename)
{
    string compile_c(string base)
    {
	just_one_compilation();
	string out = base + ".o";
	if (cli.just_compile && cli.output_file != "")
	    out = cli.output_file;
	compile_ksi(filename, out);
	return out;
    }   
	
    switch (split_filename(basename(filename))) {
    case (base, "g"):
	just_one_compilation();
	
	string obase = base;
	if (cli.output_file != "" && cli.just_compile) {
	    switch split_filename(cli.output_file) {
	    case (x, "xgo"):	obase = x;
	    case (x, "ksi"):	obase = x;
	    case (x, "o"):	obase = x;
	    case (_, _):
		die("don't know how to generate `" + cli.output_file + "'");
	    }
	}

	verb("compiling implementation " + filename + " to " + obase + ".xgo");
	compile_implementation(({input_filename = filename,
				 output_filename = obase + ".ksi",
				 output_xgo_filename = obase + ".xgo",
				 modulename = String::capitalize(base)}));
	if (!cli.just_make_ksi)
	    compile_ksi(obase + ".ksi", obase + ".o");

	return obase + ".xgo";
	
    case (base, "gi"):
	if (!cli.just_compile)
	    die(".gi files can only be handled with -c");
	    
	just_one_compilation();
	
	string out = cli.output_file;
	if (out == "")
	    out = base + ".xgi";
	    
	verb("compiling interface " + filename + " to " + out);
	compile_interface(({input_filename = filename,
			    output_filename = out,
			    output_xgo_filename = "",
			    modulename = String::capitalize(base)}));
	return out;
    
    case (base, "ksi"):
	return compile_c(base);

    case (base, "c"):
	return compile_c(base);
	
    case (_, "xgo"):
	return filename;
	
    case (_, "xga"):
	return filename;
	
    case (_, "so"):
	return filename;
	
    case (_, "a"):
	return filename;
	
    case (_, "o"):
	return filename;
	
    case (_, _):
	prerr_endline("don't know what to do with: " + filename);
	raise Bail_out;
    }
}

void run_actions()
{
    cli.lib_path = List::concat(cli.lib_path, List::cons1(cli.prefix + "/lib"));
    
    Module::add_to_search_path(cli.prefix + "/lib");
    List::iter(Module::add_to_search_path, cli.module_path);

    if (cli.ciface) {
	switch cli.input_files {
	case Cons(name, Nil):
	    Ciface::dump_ciface(name);
	    
	case _:
	    if (List::length(cli.input_files) != 1)
		die("-ciface takes exactly one argument");
	}

	return;
    }

    // put input_files in command line order, and prepend standard library
    if (!cli.just_compile && !cli.mkarch_mode)
	cli.input_files = Cons("core.xga",
			  Cons("std.xga", List::rev(cli.input_files)));
    else
	cli.input_files = List::rev(cli.input_files);
    
    cli.input_files = List::map(compile_file, cli.input_files);

    if (cli.just_compile)
	return;
    
    string aux_dir = dirname(cli.output_file);
    if (aux_dir != "")
	aux_dir = aux_dir + "/";
	
    Linker::cmdline l = ({Linker::search_path = cli.lib_path,
			  Linker::input_files = cli.input_files,
			  Linker::output_file = cli.output_file,
			  Linker::link_flags = cli.link_flags,
			  Linker::gont_link_aux = aux_dir + "gont-link-aux"});
    
    if (cli.mkarch_mode) {
	if (l.output_file == "")
	    l.output_file = "lib.xga";
	verb("making libarary of " + String::concat(" ", l.input_files) +
	     " to " + l.output_file);
	Linker::mklib(l);
    } else {
	if (l.output_file == "")
	    l.output_file = "a.out";
	verb("linking " + String::concat(" ", l.input_files) +
	     " to " + l.output_file);
	let (ld_flags, files) = Linker::link(l) in {
	    compile_ksi(l.gont_link_aux + ".ksi", l.gont_link_aux + ".o");
	    sh(cli.ksi_compiler + " -o " + l.output_file + " " +
	       String::concat(" ", files) + " " + ld_flags);
	}
    }
}

void version()
{
    print_string("gontc, version " + Config::version + " (experimental)\n");
    Sys::exit(0);
}

void usage()
{
    string msg =
"Following options are accepted:\n"
"  -c             just compile, don't link\n"
"  -S             just generate .ksi file, don't compile it\n"
"  -a             operate in make-archive mode\n"
"  -ciface        generate C interface file for given module\n"
"  -v             be verbose (can be repeated for more verbosity)\n"
"  -o FILE        set output file name to FILE\n"
"  -quote ARG     interpret ARG as ordinary file, not an option\n"
"  -L DIR         add DIR to library search path\n"
"  -I DIR         add DIR to module search path\n"
"  -ksi OPT       add OPT to options passed to Ksi compiler\n"
"  -pg            same as -ksi -pg\n"
"  -O*            same as -ksi -O*\n"
"  -f*            same as -ksi -f*\n"
"  -g*            same as -ksi -g*\n"
"  -m*            same as -ksi -m*\n"
"  -ld OPT        pass OPT to linker\n"
"  -ksicc CC      set Ksi compiler to CC\n"
"  -prefix DIR    set installation prefix to DIR\n"
"  -B DIR         same as -prefix\n"
"  -help          print this message and exit\n"
"  --help         same as -help\n"
"  -version       print version information and exit\n"
"  --version      same as -version\n"
"  -Ftime-report  dump information about time used by different passes and\n"
"                 total memory consumptions.\n"
"\n"
"Warning flags, each flag can be prefixed with \"no-\" to reverse its meaning:\n"
"  -Wall                   all warnings\n"
"  -Wnone                  same as -Wno-all\n"
"  -Werror                 treat warnings as errors\n"
"  -Wverbose               make error messages more verbose\n"
"\n"
"Warn about:\n"
"  -Wdead                  unreachable code\n"
"  -Wlabeled               break/continue without label in labeled loop\n"
"  -Wpattern               not exhaustive/redundant pattern matching\n"
"  -Wcapitalization        capitalized pattern variables\n"
"  -Wignored-value         ignoring value of expression (this excludes value\n"
"                          of `=', `++', `--' and function calls)\n"
"  -Wignored-retval        ignoring return value of function\n"
"\n"
"Warning flags default to -Wall -Wverbose -Wno-ignored-retval.\n"
"\n"
"Examples:\n"
"  gontc -c foo.gi\n"
"  gontc -c foo.g\n"
"  gontc -o bar xml.xga foo.xgo baz.xgo\n"
"  gontc -o bar xml.xga widgets.xga:force foo.xgo baz.xgo\n"
"\n"
"Report bugs to <gont@pld-linux.org>.\n";

    print_string(msg);
    Sys::exit(0);
}

cmdline process_options()
{
    cmdline c = ({
	input_files = Nil,
	output_file = "",
	ksi_compiler = Config::ksic,
	ksi_flags = "",
	link_flags = "",
	prefix = Config::prefix,

	lib_path = List::cons1("."),
	module_path = List::cons1("."),

	just_make_ksi = false,
	just_compile = false,
	mkarch_mode = false,
	ciface = false,

	verbose = 0
    });
    int err_cnt = 0;

    int argp = 1;

    string shift()
    {
	if (argp == Array::length(Sys::argv)) {
	    report_error("nothing to shift (after `" + 
			 Array::get(Sys::argv, argp - 1) + "')");
	    bye();
	}

	return Array::get(Sys::argv, argp++);
    }

    void ksi_flag(string s)
    {
	c.ksi_flags = c.ksi_flags + " " + s;
    }
    
    void warning_flag(string flag)
    {
	bool state;
	string base;

	if (flag == "-Wnone")
	    flag = "-Wno-all";
	
	if (String::length(flag) > 5 && String::sub(flag, 0, 5) == "-Wno-") {
	    base = String::sub(flag, 5, String::length(flag) - 5);
	    state = false;
	} else {
	    base = String::sub(flag, 2, String::length(flag) - 2);
	    state = true;
	}
	
	switch base {
	case "error":		flags.werror = state;
	case "shadow":		flags.wshadow = state;
	case "dead":		flags.wdead = state;
	case "labeled":		flags.wlabeled = state;
	case "pattern":		flags.wpattern = state;
	case "capitalization":	flags.wcapitalization = state;
	case "ignored-value":	flags.wignored_value = state;
	case "ignored-retval":	flags.wignored_retval = state;
	case "verbose":		flags.verbose_errors = state;
	case "all":
	    flags.wshadow = state;
	    flags.wdead = state;
	    flags.wlabeled = state;
	    flags.wpattern = state;
	    flags.wcapitalization = state;
	    flags.verbose_errors = state;
	    flags.wignored_value = state;
	    flags.wignored_retval = state;
	case _:
	    report_error("unknown warning flag: `" + flag + "'");
	    err_cnt++;
	}
    }

    void ordinary_file(string f)
    {
	c.input_files = Cons(f, c.input_files);
    }
    
    while (argp < Array::length(Sys::argv)) {
	string arg = shift();

	switch arg {
	case "-ksicc":	    c.ksi_compiler = shift();
	case "-ksi":	    ksi_flag(shift());
	case "-pg":	    ksi_flag(arg);

	case "-B":
	case "-prefix":	    c.prefix = shift();
	
	case "-ld":	    c.link_flags = c.link_flags + " " + shift();
	case "-I":	    c.module_path = Cons(shift(), c.module_path);
	case "-L":	    c.lib_path = Cons(shift(), c.lib_path);

	case "-c":	    c.just_compile = true;
	case "-a":	    c.mkarch_mode = true;
	case "-v":	    c.verbose++;
	case "-S":	    
	    c.just_make_ksi = true;
	    c.just_compile = true;
	
	case "-ciface":
	    c.ciface = true;
	
	case "-Ftime-report":
	    flags.time_report = true;

	case "-o":	    c.output_file = shift();
	
	case "-quote":	    ordinary_file(shift());

	case "-help":	    usage();
	case "--help":	    usage();

	case "-version":    version();
	case "--version":   version();

	case _:
	    if (String::length(arg) >= 2 && String::sub(arg, 0, 1) == "-")
		switch String::sub(arg, 0, 2) {
		case "-g":    ksi_flag(arg);
		case "-O":    ksi_flag(arg);
		case "-f":    ksi_flag(arg);
		case "-m":    ksi_flag(arg);
		case "-W":    warning_flag(arg);
		case _:
		    report_error("unknown knob: `" + arg + "'");
		    err_cnt++;
		}
	    else
		ordinary_file(arg);
	}
    }

    if (List::length(c.input_files) == 0) {
	report_error("no input file names passed");
	err_cnt++;
    }
    
    if (err_cnt != 0)
	bye();

    return c;
}

void main()
{
    try {
	cli = process_options();
	run_actions();
    } with {
    case Invalid_argument(m):
	prerr_endline("invalid argument exception: " + m);
	Sys::exit(1);
	
    case Failure(m):
	prerr_endline("failure exception: " + m);
	Sys::exit(1);
    
    case Sys_error(m):
	prerr_endline("syserror exception: " + m);
	Sys::exit(1);
	
    case Not_found:
	prerr_endline("not found exception");
	Sys::exit(1);
    
    case Match_failure:
	prerr_endline("match failure exception");
	Sys::exit(1);
    
    case Bail_out:
	prerr_endline("compliation terminated (Bail_out)");
	Sys::exit(1);
	
    case Choke:
	prerr_endline("compliation terminated (Choke)");
	Sys::exit(1);
    }

    Sys::exit(0);
}

section init 
{
    main();
}

// vim:sts=4:sw=4
