/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 703 $ $LastChangedDate: 2002-08-30 01:19:46 +0200 (Fri, 30 Aug 2002) $
 *
 * Util - Various utility functions: diagnostics, compilation flags, etc.
 *
 * Maybe we should just leave types in Ast and move functions here?
 */
 
open Ast;
open Util;

/*** Flags. ***/
compiler_flags flags;

section init 
{
	flags = ({
		werror = false,
		wshadow = true,
		wdead = true,
		wlabeled = true,
		wpattern = true,
		wcapitalization = true,
		verbose_errors = true,
		wignored_value = true,
		wignored_retval = false,
		time_report = false
	});
}

/*** Diagnostics. ***/
<location>Stack::t loc_stack;
bool bailout;
location last_location;

section init 
{
	loc_stack = Stack::create();
	last_location = ({
			file = "unknown", 
			first_line = 0,
			first_column = 0,
			last_line = 0,
			last_column = 0
			});
	bailout = false;
}

void message(location loc, string msg)
{
	string beg = loc.file + ":" + 
		     itoa(loc.first_line) + ":" +
		     itoa(loc.first_column);

	if (loc.last_line != 0 &&
	    (loc.last_line != loc.first_line ||
	     (loc.last_column != 0 && loc.last_column != loc.first_column)))
	     	beg = beg + "-" + itoa(loc.last_line) + ":" + 
		      itoa(loc.last_column);
	
	prerr_endline(beg + ": " + msg);
}

void warnl(location loc, string msg)
{
	message(loc, "warning: " + msg);
	if (flags.werror)
		bailout = true;
}

void errorl(location loc, string msg)
{
	message(loc, msg);
	bailout = true;
}

/* Location stack. */
void push_location(location loc)
{
	last_location = loc;
	Stack::push(loc_stack, loc);
}

void pop_location()
{
	Stack::pop(loc_stack);
}

location top_location()
{
	if (Stack::empty(loc_stack))
		return last_location;
	else
		return Stack::top(loc_stack);
}

void warn(string msg)
{
	warnl(top_location(), msg);
}

void error(string msg)
{
	errorl(top_location(), msg);
}

void maybe_choke()
{
	if (bailout == true)
		raise Bail_out;
}

'a locate(*('a ('b)) f, <'b>located x)
{
	'a r;

	push_location(x.loc);
	try {
		r = f(x.a);
	} finally {
		pop_location();
	}

	return r;
}

'a ice(string msg)
{
	if (bailout == true) {
		error("confused by earlier errors, bailing out [" + msg + "]");
		raise Bail_out;
	}
	
	error("internal compiler error: " + msg);
	return Sys::abort();
}


/*** General utils. ***/
'a get_binding(<'a>bindable sym)
{
	switch sym {
	case Symbol(x):
		return x;
	case SimpleName(n):
		return ice("get_binding(): symbol '" + n + "' not bound");
	case QualifiedName(m, n):
		return ice("get_binding(): qualified symbol '" + 
			   m + "::" + n + "' not bound");
	}
}

void dbg(string s)
{
	print_string(s);
}

void dbgl(string s)
{
	message(top_location(), "DEBUG: " + s);
}

bool same_symbols(<'a>symbol s1, <'a>symbol s2)
{
	return s1.module == s2.module && s1.name == s2.name;
}

// TODO: move to Std?
*('a ('b)) partial(*('a ('c, 'b)) f, 'c c)
{
	return (fun 'a ('b b) (f(c, b)));
}

*('a ('b)) partial2(*('a ('c, 'd, 'b)) f, 'c c, 'd d)
{
	return (fun 'a ('b b) (f(c, d, b)));
}

// Unique name generator

struct ung_counter {
	int cnt;
}

<ung_counter>Hashtbl::t ung_counters;

section init
{
	ung_counters = Hashtbl::create(13);
}

string genname(string prefix)
{
	int cnt;
	
	if (Hashtbl::mem(ung_counters, prefix)) {
		let c = Hashtbl::find(ung_counters, prefix) in
		cnt = c.cnt++;
	} else {
		cnt = 1;
		Hashtbl::add(ung_counters, prefix, ({ cnt = 2 }));
	}

	return prefix + "$" + itoa(cnt);
}

// Ast manipulation

string simple_name(<<'a>symbol>bindable s)
{
    switch s {
    case SimpleName(x):
	return x;
	
    case QualifiedName(_, x):
	return x;
	
    case Symbol(x):
	return x.name;
    }
}

string symbol_name(<'a>symbol x)
{
    if (x.module != "")
	return x.module + "::" + x.name;
    else
	return x.name;
}

string qualified_name(<<'a>symbol>bindable s)
{
    switch s {
    case SimpleName(x):
	return x;
	
    case QualifiedName(m, n):
	return m + "::" + n;
	
    case Symbol(x):
	return symbol_name(x);
    }
}

<<'a>symbol>bindable rev_qualified_name(string name)
{
    <string>list l = String::split("::", name);
    
    switch l {
    case Cons(m, Cons(n, Nil)):
	return QualifiedName(m, n);
    case Cons(n, Nil):
	return SimpleName(n);
    case _:
	raise Invalid_argument("Ast::rev_qualified_name");
    }
}

void set_property(var_symbol s, var_symbol_property p)
{
    s.a.properties = Cons(p, s.a.properties);
}

bool has_property(var_symbol s, var_symbol_property p)
{
    return List::exists((fun bool (var_symbol_property x) 
		  		(Hacks::union_tag(x) == Hacks::union_tag(p))),
			s.a.properties);
}

ty type_map(*(ty (ty)) f, ty t)
{
    ty self(ty tt)
    {
	return type_map(f, tt);
    }

    <'a>field self_fld(<'a>field fld)
    {
	ty tt;

	push_location(fld.fld_loc);
	try {
	    tt = self(fld.fld_type);
	} finally {
	    pop_location();
	}
	
	return ({fld_type = tt,
	     fld_symbol = fld.fld_symbol,
	     fld_loc = fld.fld_loc});
    }
    
    t = f(t);
    
    switch t {
    case Tuple(lst):
	return Tuple(List::map(self, lst));
    
    case Func(ft):
	*(ty, <id>option) map_fnc_arg(*(ty, <id>option) p)
	{
	    let (tt, nn) = p in
	    return (self(tt), nn);
	}
	
	return Func(({ret_type = self(ft.ret_type),
		      parms = List::map(map_fnc_arg, ft.parms),
		      tyvars = ft.tyvars}));
    
    case Struct(flds):
	return Struct(List::map(self_fld, flds));
	
    case OptStruct(flds):
	return OptStruct(List::map(self_fld, flds));
	
    case Union(flds):
	return Union(List::map(self_fld, flds));
    
    case NamedType(nt):
	return NamedType(({parms = List::map(self, nt.parms),
		   	   tydef = nt.tydef}));

    case XTyVar(_):
	return t;   // ???

    case _:
	return t;
    }
}

ty copy_type(ty t)
{
    ty f(ty tt)
    {
	switch tt {
	case XTyVar(x):
	    _ xx = {t = None, name = x.name + "c", flag = x.flag};
	    switch x.t {
	    case Some(y):
		xx.t = Some(copy_type(y));
	    case None:
		skip;
	    }
	    return XTyVar(xx);
	case _:
	    return tt;
	}
    }

    return type_map(f, t);
}

void type_iter(*(void (ty)) f, ty t)
{
    void self(ty tt)
    {
	type_iter(f, tt);
    }

    void self_fld(<'a>field fld)
    {
	push_location(fld.fld_loc);
	try {
	    self(fld.fld_type);
	} finally {
	    pop_location();
	}
    }
    
    f(t);
    
    switch t {
    case Tuple(lst):
	List::iter(self, lst);
    
    case Func(ft):
	void map_fnc_arg(*(ty, <id>option) p)
	{
	    let (tt, _) = p in
	    self(tt);
	}
	
	self(ft.ret_type);
	List::iter(map_fnc_arg, ft.parms);
    
    case Struct(flds):
	List::iter(self_fld, flds);
	
    case OptStruct(flds):
	List::iter(self_fld, flds);
	
    case Union(flds):
	List::iter(self_fld, flds);
    
    case NamedType(nt):
	List::iter(self, nt.parms);
	
    case XTyVar(xt):
	switch xt.t {
	case Some(tt):
	    self(tt);
	case None:
	    skip;
	}
    
    case _:
	skip;
    }
}
